#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "tokenizer/File.h"
#include "utils/Utils.h"

#define MAX_DEFINES 50

static FileError fileError = NULL;
static FILE* file = NULL;
static int readIndex = 0;
static int writeIndex = 0;
static int fileSize = 0;
static char* fileContent = NULL;

typedef struct {
    int nameLength;
    const char* name;
    const char* code;
} Defines;

static int defineIndex = 0;
static Defines defines[MAX_DEFINES];
static int defineReadIndex = 0;
static const char* defineCode = NULL;

static void fAdd(int c) {
    if(writeIndex >= fileSize) {
        fileSize *= 2;
        fileContent = realloc(fileContent, fileSize);
    }
    fileContent[writeIndex++] = c;
}

static void fReset() {
    readIndex = 0;
    writeIndex = 0;
    fileSize = 16;
    fileContent = malloc(fileSize);
    defineIndex = 0;
    defineReadIndex = 0;
    defineCode = NULL;
}

void fOpen(const char* path, FileError fe) {
    fileError = fe;
    file = fopen(path, "r");
    if(file == NULL) {
        fileError("cannot read file '%s'", path);
        return;
    }
    fReset();
    while(true) {
        int c = fgetc(file);
        if(c == '#') {
            while(c != EOF && c != '\n') {
                fAdd(c);
                c = fgetc(file);
            }
            fAdd('\0');
        }
        fAdd(c);
        if(c == EOF) {
            break;
        }
    }
    fAdd('\0');
    fclose(file);
}

void fClose() {
    free(fileContent);
}

static void fSkipString(int skip) {
    readIndex += skip;
    while(fileContent[readIndex] == ' ') {
        readIndex++;
    }
}

static void fDefine() {
    fSkipString(7);
    const char* name = fileContent + readIndex;
    while(true) {
        if(fileContent[readIndex] == ' ') {
            fileContent[readIndex] = '\0';
            break;
        } else if(!isLetter(fileContent[readIndex])) {
            fileError("invalid define name '%s'", name);
        }
        readIndex++;
    }
    readIndex++;
    if(defineIndex >= MAX_DEFINES) {
        fileError("too many defines");
    }
    defines[defineIndex].nameLength = strlen(name);
    defines[defineIndex].name = name;
    defines[defineIndex].code = fileContent + readIndex;
    defineIndex++;
    fSkipString(strlen(fileContent + readIndex));
}

static bool fMatch(Defines* def, const char* code) {
    return strncmp(def->name, code, def->nameLength) == 0;
}

static bool fPrepareChar() {
    if(defineCode != NULL) {
        if(defineCode[defineReadIndex] == '\0') {
            defineCode = NULL;
            defineReadIndex = 0;
        } else {
            return true;
        }
    }
    if(!isLetter(fileContent[readIndex])) {
        defineReadIndex = 0;
    }
    if(isLetter(fileContent[readIndex]) && defineReadIndex == 0) {
        const char* replace = fileContent + readIndex;
        for(int i = 0; i < defineIndex; i++) {
            if(fMatch(defines + i, replace)) {
                defineCode = defines[i].code;
                readIndex += defines[i].nameLength;
                return true;
            }
        }
        defineReadIndex = -1;
    }
    if(fileContent[readIndex] == '#') {
        const char* command = fileContent + readIndex + 1;
        if(strncmp(command, "define", 6) == 0) {
            fDefine();
        } else {
            fileError("unknown preprocessor command '%s'", command);
        }
    }
    if(fileContent[readIndex] == '\0') {
        readIndex++;
    }
    return false;
}

int fRead() {
    if(fPrepareChar()) {
        return defineCode[defineReadIndex++];
    }
    return fileContent[readIndex++];
}

int fPeek() {
    if(fPrepareChar()) {
        return defineCode[defineReadIndex];
    }
    return fileContent[readIndex];
}

bool fReadIf(int c) {
    if(fPeek() == c) {
        fRead();
        return true;
    }
    return false;
}